%%
% X: data input
% D: left dictionary
% R: right dictionary
% T0: percentage of total atoms to select
% katoms: number of atoms per iteration
% powers: hop neibors
% groups: number of groups 
% solver: Y,W solver options

function [Ys, Ws, selectedIss, selectedJss, times]=LRMDS(X, D, R, T0, katoms, powers, groups, solver)

k = katoms;
p = powers;
g = groups;
[N,SN]=size(D);
[SM,M]=size(R);

tol = 1e-6 * N*M;

% TGSD params
parm.K = g;
parm.lambda_1 = .01;
parm.lambda_2 = .01;
parm.lambda_3 =1;

parm.rho_1=parm.lambda_1;
parm.rho_2=parm.lambda_2;
parm.iter=500;
parm.mask =[];
parm.Psi_orth=1;
parm.Phi_orth=0;
type='rand';


R = R';

Re = X;
Re(isnan(Re)) = 0;
norm_pre = sum(Re.^2, 'all');

Dn = D ./ vecnorm(D);
Rn = R ./ vecnorm(R);

selected_linear = [];
selectedIs = [];
selectedJs = [];

Dnp = Dn.^p;
Rnp = Rn.^p;
for t = 1 : SN + SM

    st = tic;
    
    projs = (Dnp)'*(Re.^p)*(Rnp);
    resMatrix = abs(projs);
    resMatrix(selected_linear) = nan;
    
    [val_sorted, idx_sorted] = sort(resMatrix(:), "descend");
    ist = find(~isnan(val_sorted), 1);

    if isnan(val_sorted(ist))
        break;
    end

    idx_sorted_valid = idx_sorted(ist:end);
    idl = length(idx_sorted_valid);

    ct = 0;
    s = 1;
    idone = 0;
    jdone = 0;
    while ct < k && s < idl
        
        ij_max = idx_sorted_valid(s);
        [i_max, j_max] = ind2sub([SN SM], ij_max);

        if length(selectedIs) > SN
            idone = 1;
        elseif ~ismember(i_max, selectedIs)
            selectedIs(end+1) = i_max;
            ct = ct+1;
        end
        
        if length(selectedJs) > SM
            jdone = 1;
        elseif ~ismember(j_max, selectedJs) 
            selectedJs(end+1) = j_max;
            ct = ct+1;
        end
        
        if idone && jdone
            break;
        end
        s = s + 1;
    end

    if idone && jdone
        break;
    end

    time_find_best(t) = toc(st);

    % solver Y, W
    st = tic;

    Dt = D(:,selectedIs);
    Rt = R(:,selectedJs);

    selectedIss{t} = selectedIs;
    selectedJss{t} = selectedJs;
    
    if solver == 1 % X = DYWR
        sz = min([g, length(selectedIs), length(selectedJs)]);
        Yt = rand(size(Dt,2), sz);
        prt = pinv(Rt');
        pdt = pinv(Dt);
        for it = 1:200
            Wt = pinv(Dt * Yt) * X * prt;
            Y_it = pdt * X * pinv(Wt * Rt');
    
            if norm(Y_it-Yt,"fro") < 1e-6 * size(Dt,2) * g
                disp("converged at " + it);
                Yt = Y_it;
                break;
            end
            
            Yt = Y_it;

        end
    elseif solver == 2 % X = DZR, Z = YW   
        Z = pinv(Dt) * X * pinv(Rt');
        [U, S, V] = svd(Z,"econ");
        sz = min(size(S,1),g);
        Yt = U(:,1:sz);
        Wt = S(1:sz,1:sz)*V(:,1:sz)';
    elseif solver ==3
        Z = inv(Dt'*Dt) * Dt' * X * Rt * inv(Rt' * Rt);
        [U, S, V] = svds(Z, g);
        Yt = U;
        Wt = S*V';
        sz = min(size(S,1),g);
    else % TGSD
        sz = g;
        [~,Y0,Sigma,W0,Wt,Yt]= TGSD_optimization(X,Dt,Rt',parm,type);
    end

    time_solve_YW(t) = toc(st);

    st = tic;
    Y = zeros(SN,sz);
    W = zeros(sz,SM);
    
    Y(selectedIs,:) = Yt;
    W(:,selectedJs) = Wt;

    Ys{t} = Y;
    Ws{t} = W;

    Xt = Dt * Yt * Wt * Rt';
    Re = X - Xt; 

    time_re_update(t) = toc(st);

    norm_curr = sum(Re.^2, 'all');

    if  norm_curr < tol || abs(norm_curr-norm_pre) / norm_pre < 1e-6
        disp("Residual converged !!!");
        break;
    end

    norm_pre = norm_curr;
    disp("iter - " + t + ": " + norm_curr);
    
    if (length(selectedIs) + length(selectedJs)) >= T0 * (SM + SN)
        disp("target atoms count meet.")
        break;
    end
end

times = [time_find_best;
    time_solve_YW;
    time_re_update];

end
